/* 
 * Mobile browser shell
 */
/* APIs:
	nsIXULBrowserWindow: setJSStatus, setOverLink
	nsIWebProgressListener: state, location, status, security listener
	nsIBrowserDOMWindow: The C++ source has access to browser script through this
				openURI, isTabContentWindow
*/
const nsIBrowserDOMWindow		= Components.interfaces.nsIBrowserDOMWindow;
const nsIDOMChromeWindow		= Components.interfaces.nsIDOMChromeWindow;
const nsIDOMWindow			 	= Components.interfaces.nsIDOMWindow;
const nsIDocShellHistory		= Components.interfaces.nsIDocShellHistory;
const nsIDocShellTreeItem		= Components.interfaces.nsIDocShellTreeItem;
const nsIInterfaceRequestor		= Components.interfaces.nsIInterfaceRequestor;
const nsISHistory				= Components.interfaces.nsISHistory;
const nsISupportsWeakReference 	= Components.interfaces.nsISupportsWeakReference;
const nsIWebNavigation		 	= Components.interfaces.nsIWebNavigation;
const nsIXULBrowserWindow		= Components.interfaces.nsIXULBrowserWindow;
const nsIWebProgress			= Components.interfaces.nsIWebProgress;
const nsIWebProgressListener	= Components.interfaces.nsIWebProgressListener;
const nsIWebProgressListener2	= Components.interfaces.nsIWebProgressListener2;
const nsIXULWindow			 	= Components.interfaces.nsIXULWindow;
const nsIConsoleService			= Components.interfaces.nsIConsoleService;
const nsIObserverService		= Components.interfaces.nsIObserverService;

function debugBrowser() {
	var browser = getShell().browser;
	var body = browser.contentWindow.document.body;
	var html = browser.contentWindow.document.documentElement;
	var scale = browser.markupDocumentViewer.fullZoom;
	dump(browser.currentURI.spec + " scale" + scale + "\n");
	dump("body clientWidth:" + body.clientWidth + " scrollWidth:" + body.scrollWidth + "\n");
	dump("html clientWidth:" + html.clientWidth + " scrollWidth:" + html.scrollWidth + "\n");
	dump("scrollX: " + browser.contentWindow.scrollX + "\n");
}



function ZoomController(aBrowser) {
	this.browser = aBrowser;
};

// ZoomControler sets browser zoom
// It clamps/
ZoomController.prototype = {
	minScale: 0.1,
	maxScale: 3,
	set scale(s) 
	{
		var clamp = Math.min(this.maxScale, Math.max(this.minScale, s));
		clamp = Math.floor(clamp * 1000) / 1000;	// Round to 3 digits
		if (clamp == this.browser.markupDocumentViewer.fullZoom)
			return;
//		dump("scrollY: " + this.browser.contentWindow.scrollY + "\n");//
//		dump("scaling to: " + clamp + "\n");
		this.browser.markupDocumentViewer.fullZoom = clamp;
//		dump("after scrollY: " + this.browser.contentWindow.scrollY + "\n");
		// if we've zoomed out of the viewport, scroll us back in
		var leftEdge = this.browser.contentWindow.scrollX + this.browser.contentWindow.document.documentElement.clientWidth;
		var scrollX = this.browser.contentWindow.document.documentElement.scrollWidth - leftEdge;
 //	 dump("leftEdge:" + leftEdge + " maxWidth:" + this.browser.contentWindow.document.body.style.maxWidth + "\n");
		if (scrollX < 0)
		{
//			dump("had to scroll\n");
			this.browser.contentWindow.scrollBy( scrollX, 0);
		}
	},
	get scale()
	{
		return this.browser.markupDocumentViewer.fullZoom;
	},
	reset: function() 
	{
		this.minScale = ZoomController.prototype.minScale;
		this.maxScale = ZoomController.prototype.maxScale;
	},
	fitContent: function(animate) 
	{
//		dump("fitContent\n");
		animate = animate || false;
		this.scale = 1;		// reset the scale to 1 forces document to preferred size
		var body = this.browser.contentWindow.document.body;
		var html = this.browser.contentWindow.document.documentElement;
		var newScale = this.scale;
		var finalWidth = html.clientWidth;
		var prefScrollWidth = Math.max(html.scrollWidth, body.scrollWidth); // empirical hack, no idea why
//dump("Browser width:" + this.browser.boxObject.width + "\n");
//dump("WIDTHS: " + "scroll: " + html.scrollWidth + " client " + html.clientWidth + " final:" + finalWidth + "preferred: " + prefScrollWidth + "\n");
		if (prefScrollWidth > (this.browser.boxObject.width - 10) )	{ 
			// body wider than window, scale id down
			// we substract 10 to compensate for 10 pixel browser left margin
			newScale = (this.browser.boxObject.width ) / prefScrollWidth;
			finalWidth = prefScrollWidth;
//			dump("scaling down to " + finalWidth + " scale: " + newScale + "\n");
		}
		body.style.minWidth = body.style.maxWidth = (finalWidth -20) + "px";
		this.minScale = Math.max(this.minScale, newScale);
		if (animate) {
			var zoomController = this;
			var i = Math.max(Math.abs(this.scale - newScale) / 0.1, 1);
			var scaleStep = (newScale - this.scale) / i;
			var intervalId = window.setInterval( function() {
				if (i > 0)
					zoomController.scale += scaleStep;
				else {
					zoomController.scale = newScale;
					window.clearInterval(intervalId);
				}
				i -= 1;
			}, 0);
		}
		else
			this.scale = newScale;
//	dump("scale: " + newScale + " finalWidth: " + finalWidth + "\n");
	},
	getPagePosition: function (el) 
	{
		// this function is often wrong. We really need browser support.  getBoxObjectFor was promising,
		// but it was buggy when zoomed.
		//		var box = el.ownerDocument.getBoxObjectFor(el);
		//		var retVal = { left: box.x, top: box.y, width: box.width, height: box.height,
		//			screenX: box.screenX, screenY: box.screenY};
		var top = 0, left = 0;
		var retVal = { width: el.offsetWidth, height: el.offsetHeight, y:0, x:0};
		while (el != null) {
			retVal.y += el.offsetTop;
			retVal.x += el.offsetLeft;
			el = el.offsetParent;
		}
//		dump("pagePosition x:" + retVal.x +  " y:" + retVal.y + " width:" + retVal.width + " height:" + retVal.height + "\n");
		return retVal;
	},
	getWindowRect: function() 
	{
		return { x: this.browser.contentWindow.scrollX,
				y: this.browser.contentWindow.scrollY,
				width: this.browser.boxObject.width / this.scale,
				height: this.browser.boxObject.height / this.scale }
	},
	toggleZoom: function(el) 
	{
		if (!el) return;
		var elRect = this.getPagePosition(el);
		var winRect = this.getWindowRect();
		// Element is already revealed if:
		if (elRect.x > winRect.x // left is visible
			&& (elRect.x + elRect.width) < (winRect.x + winRect.width) // right is visible
			&& elRect.y > winRect.y // top is visible
			&& elRect.y < (winRect.y + winRect.height * 0.5) // we are in top half of screen
			&& elRect.width > winRect.width * 0.8 ) {	// element is wider than 80% of screen	
			this.fitContent(true);	// already revealed, zoom out
		}
		else
			this.zoomToElement(el);
	},
	getZoomDestination: function(el) 
	{
		var elRect = this.getPagePosition(el);
		var winRect = this.getWindowRect();
		// destination: slightly bellow the top, and zoomed in so you can see element's surroundings
		var margin = elRect.width / 12;
		var x = elRect.x;
		var width = elRect.width;
		if (elRect.width + margin * 2 < winRect.width) {
			x = Math.max( x - margin, 0);
			width += 2* margin;
		}
		return { x: x, 
				y: Math.max(0, elRect.y - winRect.height * 0.4), 
				scale: (this.browser.boxObject.width- 20) / width };
	},
	zoomToElement: function(el) 
	{	
		var dest = this.getZoomDestination(el);
		var distance = Math.max(Math.abs(dest.y - this.browser.contentWindow.scrollY), 
								Math.abs(dest.x - this.browser.contentWindow.scrollX));
		var scaleDist = Math.abs(this.scale - dest.scale);
		var stepCount = Math.max(distance/20, scaleDist / 0.1); // step count depends on distance in zoom/scrolling
		var stepCount = Math.min(10, Math.floor(stepCount));
		var zoomController = this;
		var intervalId = window.setInterval( function() {	
			dest = zoomController.getZoomDestination(el);
//			dump("stepCount:" + stepCount + "\n"); 
//			dump("zoomin to x:" + dest.x + " y:" + dest.y + " scale: " + dest.scale + "\n");
			if (stepCount <= 0)	{
				zoomController.scale = dest.scale;
				zoomController.browser.contentWindow.scrollTo(dest.x, dest.y);
				window.clearInterval(intervalId);				
			}
			else {
				var x = Math.floor(zoomController.browser.contentWindow.scrollX + (dest.x - zoomController.browser.contentWindow.scrollX) / stepCount);
				var y = Math.floor(zoomController.browser.contentWindow.scrollY + (dest.y - zoomController.browser.contentWindow.scrollY) / stepCount);
				var scale = zoomController.scale + (dest.scale - zoomController.scale) / stepCount;
//				dump("slide to x:" + x + " y:" + y + " scale:" + scale + 
//				" scrollY:" + zoomController.browser.contentWindow.scrollY + "\n");
				zoomController.scale= scale;
				zoomController.browser.contentWindow.scrollTo(x, y);
			}
			stepCount -= 1;
		}, 0);
	}
};


var gShell = null;
function getShell() {
	if (gShell == null)
		gShell = new TouchShell();
	return gShell;
}

var TouchShell = function() {}

// TouchShell: mobile application shell
// 
TouchShell.prototype = {
	init: function () 
	{
		try {
			this.browser = document.getElementById("content");
			this.browser.zoomController = new ZoomController(this.browser);
			this.browser.mouseController = new MouseController(this.browser);
			this.browser.addEventListener("pageshow", this, true);

			// Lots of boilerplate application initialization code from here on
			this.listener = new nsBrowserStatusHandler();
			window.XULBrowserWindow = this.listener; 
			window.QueryInterface(nsIInterfaceRequestor)
				.getInterface(nsIWebNavigation)
				.QueryInterface(nsIDocShellTreeItem).treeOwner
				.QueryInterface(nsIInterfaceRequestor)
				.getInterface(nsIXULWindow)
				.XULBrowserWindow = window.XULBrowserWindow;
		
			window.QueryInterface(nsIDOMChromeWindow).browserDOMWindow = new nsBrowserAccess();
			var appCore = Components.classes["@mozilla.org/appshell/component/browser/instance;1"].createInstance(Components.interfaces.nsIBrowserInstance);	
			appCore.setWebShellWindow(window);
		
			this.browser.webNavigation.sessionHistory = 
				Components.classes["@mozilla.org/browser/shistory;1"].createInstance(nsISHistory);
			this.browser.removeAttribute("disablehistory");	
			this.browser.docShell.QueryInterface(nsIDocShellHistory).useGlobalHistory = true;
			
			this.browser.addProgressListener(this.listener, nsIWebProgress.NOTIFY_ALL);
			var prefs = Components.classes['@mozilla.org/preferences-service;1'].getService(Components.interfaces.nsIPrefBranch);
			this.loadURI(prefs.getCharPref("browser.startup.homepage"));
		} 
		catch (e) {
			alert(e);
		}
	},
	handleEvent: function(e) 
	{
		switch(e.type) {
			case "pageshow":
//				dump("pageshow");
				this.browser.zoomController.fitContent();
				break;
			default:
				dump("Unknown event type: " + e.type + "\n");
				break;
		}
	},
	loadURI: function(uri)
	{
		if (!uri) return;
		getChrome().prepareForLoad();
		this.browser.webNavigation.loadURI(uri, nsIWebNavigation.LOAD_FLAGS_ALLOW_THIRD_PARTY_FIXUP, null, null, null);	
	},
	startDocumentLoad : function(aRequest)
	{
		this.browser.zoomController.reset();
		getChrome().startLoad();
	},
	endDocumentLoad : function(aRequest, aStatus)
	{
		getChrome().endDocumentLoad();	// fix up chrome
		this.browser.contentWindow.scrollbars.visible = false;
		// dump("endDocumentLoad\n");
		this.browser.zoomController.fitContent(); // zoom content
		// initialize scrolling
		this.browser.mouseController.documentLoaded();
	},
	goBack: function()
	{ 
		if (this.browser.webNavigation.canGoBack)
			this.browser.webNavigation.goBack(); 
	},
	goForward: function()
	{
		if (this.browser.webNavigation.canGoForward)
			this.browser.webNavigation.goForward();
	},
	stopLoading: function()
	{
		this.browser.webNavigation.stop(nsIWebNavigation.STOP_ALL);
	},
	zoomOut: function()
	{
		this.browser.zoomController.scale = 0.5;
	},
	zoomIn: function()
	{
		this.browser.zoomController.scale = 2;	
	},
	reload: function()
	{
		this.browser.webNavigation.reload(nsIWebNavigation.LOAD_FLAGS_NONE);
	},
	newTab: function()
	{
		alert ("TODO");
			//window.QueryInterface(nsIDOMChromeWindow).browserDOMWindow.openURI(null,null,null,null);
	},
	addBookmark: function()
	{
		alert("TODO");
	}
}

function nsBrowserStatusHandler() {}

nsBrowserStatusHandler.prototype = 
{
	QueryInterface : function(aIID)
	{
		if (aIID.equals(nsIWebProgressListener) ||
			aIID.equals(nsIXULBrowserWindow) ||
			aIID.equals(nsISupportsWeakReference) ||
			aIID.equals(nsISupports))
			return this;
		throw Components.results.NS_NOINTERFACE;
	},
	init : function()
	{
		this.onSecurityChange(null, null, nsIWebProgressListener.STATE_IS_INSECURE);
	},	
	destroy : function()
	{
	},
	onStateChange : function(aWebProgress, aRequest, aStateFlags, aStatus)
	{
		
		if (aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
		// Certain sites (digg.com), do not give us the STATE_IS_NETWORK/STATE_STOP event
		// still working on debugging this
			if (aStateFlags & nsIWebProgressListener.STATE_START) {
				if(aRequest && aWebProgress.DOMWindow == content)
					getShell().startDocumentLoad(aRequest);
			}
			else if (aStateFlags & nsIWebProgressListener.STATE_STOP) {
				if (aRequest && aWebProgress.DOMWindow == content) 
					getShell().endDocumentLoad(aRequest, aStatus);
			}
		}
	},
	onProgressChange : function(aWebProgress, aRequest, aCurSelfProgress, aMaxSelfProgress, aCurTotalProgress, aMaxTotalProgress)
	{
		this.currentTotalProgress = aCurTotalProgress;
		this.maxTotalProgress	 = aMaxTotalProgress;

		var percentage = aCurTotalProgress/aMaxTotalProgress;
		
		//document.getElementById("progressbar").value = parseInt(percentage*100)+"%";	 
		// TODO
		if (percentage > 0.2)
			getChrome().updateTitle();
	},
	onLocationChange : function(aWebProgress, aRequest, aLocation)
	{
	},	
	onStatusChange : function(aWebProgress, aRequest, aStatus, aMessage)
	{ // TODO
	},
	onSecurityChange : function(aWebProgress, aRequest, aState)
	{ // TODO
	},
	setJSStatus : function(status)
	{ // TODO
	},
	setJSDefaultStatus : function(status)
	{ // TODO
	},
	setDefaultStatus : function(status)
	{ // TODO
	},
	setOverLink : function(link, b)
	{ // TODO
	}
}

function nsBrowserAccess() {}

nsBrowserAccess.prototype =
{
	QueryInterface : function(aIID)
	{
		if (aIID.equals(nsIBrowserDOMWindow) ||
			aIID.equals(nsISupports))
			return this;
		throw Components.results.NS_NOINTERFACE;
	},
	
	openURI : function(aURI, aOpener, aWhere, aContext)
	{
		var url = aURI ? aURI.spec : "about:blank";
		var newTab	 = getShell().browser.addTab(url);
		var newWindow	= getShell().browser.getBrowserForTab(newTab).docShell
								 .QueryInterface(nsIInterfaceRequestor)
								 .getInterface(nsIDOMWindow);
		return newWindow;
	},
	
	isTabContentWindow : function(aWindow)
	{
		var browsers = getShell().browser.browsers;
		for (var ctr = 0; ctr < browsers.length; ctr++)
			if (browsers.item(ctr).contentWindow == aWindow)
				return true;
		return false;
	}
}

var gChrome = null;

function getChrome() {
	if (gChrome == null)
		gChrome = new MobileChrome();
	return gChrome;
} 

function MobileChrome() {
	this.init();
}

// MobileChrome is an umbrella class for chrome manipulation
MobileChrome.prototype = {
	init: function() 
	{
	},
	updateTitle: function() 
	{
		document.getElementById("title").value = getShell().browser.contentTitle;
		document.getElementById("mburlbar").value = getShell().browser.currentURI.spec;
	},
	endDocumentLoad: function()
	{
		document.getElementById("load").image = "image/refresh.png"; 
		document.getElementById("load").command = "cmd_BrowserReload";
		this.updateTitle();
	},
	prepareForLoad: function()
	{
		document.getElementById('title').setAttribute('hidden','false');
		document.getElementById('urlToolBar').setAttribute('hidden','false');
		document.getElementById('urlToolBar').setAttribute('hidden','false');
		document.getElementById('load').setAttribute('hidden','false');
		document.getElementById('bm').setAttribute('hidden','false');
		document.getElementById('sysKeyboard').setAttribute('hidden','true');
		document.getElementById('commandDeck').setAttribute('selectedIndex','0');	 
		this.hideBookmarks();	
	},
	startLoad: function() 
	{
		document.getElementById("load").image = "image/cancel.png"; 
		document.getElementById("load").command = "cmd_BrowserStop"; 
		document.getElementById("title").value = "Loading...";
	},
	showBookmarks: function ()
	{
		var elem=document.getElementById('bookmarkList');
		var hiddenParm = elem.getAttribute('hidden');
		if (hiddenParm == "true")
			elem.setAttribute('hidden','false');
		else
			elem.setAttribute('hidden','true');
	},
	hideBookmarks: function ()
	{
		document.getElementById('bookmarkList').setAttribute('hidden','true');
	},
	initButtons: function()
	{
		document.getElementById('load').setAttribute('hidden','false');
		document.getElementById('title').setAttribute('hidden','false');
		document.getElementById('bm').setAttribute('hidden','false');
		document.getElementById('commandDeck').setAttribute('selectedIndex','0');
		document.getElementById('sysKeyboard').setAttribute('hidden','true');
	},
	showURLBar: function()
	{
		document.getElementById('title').setAttribute('hidden','true');
		document.getElementById('load').setAttribute('hidden','true');
		document.getElementById('bm').setAttribute('hidden','true');
		document.getElementById('commandDeck').setAttribute('selectedIndex','1');
		this.hideBookmarks();	// not clean, will be fixed
		document.getElementById('sysKeyboard').setAttribute('hidden','false');
	},
	toggleUrlToolBar: function()
	{
		var elem=document.getElementById('urlToolBar');
		if (elem.getAttribute('hidden') == "true")
			elem.setAttribute('hidden','false');
		else {
			elem.setAttribute('hidden','true');
			getChrome().hideBookmarks();
		}		
		document.getElementById('load').setAttribute('hidden','false');
		document.getElementById('bm').setAttribute('hidden','false');
		document.getElementById('commandDeck').setAttribute('selectedIndex','0');
		document.getElementById('sysKeyboard').setAttribute('hidden','true');
	},
	selectBookmark:	function(event)
	{
		var elem=document.getElementById('bookmarkList'); 
		elem.setAttribute('hidden','true');
		var item=elem.getAttribute('selectedItem');
		var itemName=item.getAttribute('label');	
	}
}


var MouseController = function(browser) {
	this.init(browser);
}

MouseController.prototype = {
	init: function(browser)
	{
		this.browser = browser;
		this.moveStyle = "none";
	},
	documentLoaded: function() 
	{
		this.browser.contentDocument.addEventListener("mousedown", this, true);
		this.browser.contentDocument.addEventListener("mouseup",this, true);
		this.browser.contentDocument.addEventListener("dblclick", this, true);
		this.moveStyle = "none";
	},
	handleEvent: function(e) 
	{
		if (! e.type in this)
			dump("MouseController called with unknown event type " + e.type + "\n");
		this[e.type](e);
	},
	mousedown: function(e) 
	{
		if (e.target instanceof HTMLInputElement)
			return;
		this.lastEvent = this.firstEvent = e;
		this.fingerDistance = 100;
		this.mousemove = e.button != 2 ? this.mousePan : this.mouseZoom;
		this.browser.contentDocument.addEventListener("mousemove", this, true);
		e.stopPropagation();
		e.preventDefault();
	},
	mouseup: function(e)
	{
		this.browser.contentDocument.removeEventListener("mousemove", this, true);
		this.moveStyle = "none";
	},
	mouseZoom: function(e)
	{
		var deltaX = e.screenX - this.firstEvent.screenX + 100;
		var deltaY = e.screenY - this.firstEvent.screenY;
		var newDist = Math.sqrt(Math.pow(deltaX, 2) + Math.pow(deltaY, 2));
		var scale = newDist / this.fingerDistance;
		if (e.screenX < this.firstEvent.screenX && scale > 1)
			scale = 1 / scale;
		var newZoom = scale * this.browser.markupDocumentViewer.fullZoom;
//		dump("dist: " + newDist + " newZoom: " + newZoom + "\n");
		this.fingerDistance = Math.max(0.1, newDist);
		this.browser.zoomController.scale = newZoom;
		this.lastEvent = e;
		e.stopPropagation();
		e.preventDefault();
	},
	mousePan: function(e)
	{
//		if (this.lastEvent)
//			dump(e.timeStamp + " event delta:" + ( e.timeStamp - this.lastEvent.timeStamp ) + "\n");
		this.browser.contentWindow.scrollBy(
			this.lastEvent.clientX - e.clientX,
			this.lastEvent.clientY - e.clientY);
		this.lastEvent = e;
		e.stopPropagation();
		e.preventDefault();
	},
	dblclick: function(e) 
	{
		// find the target by walking the dom.
		// we want to zoom in on the block elements
		var initTarget = e.explicitOriginalTarget;
		e.preventDefault();
		while (initTarget && initTarget.nodeName != "HTML")
		{
			var disp = window.getComputedStyle(initTarget, "").getPropertyValue("display");
			if (!disp.match(/(inline)/g))
			{
				this.browser.zoomController.toggleZoom(initTarget);
				break;
			}
			else
				initTarget = initTarget.parentNode;
		}
	}
}
